# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-08-11 16:52
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import django_afip.models


class Migration(migrations.Migration):

    replaces = [('afip', '0001_initial'), ('afip', '0002_auto_20150909_1837'), ('afip', '0003_vat_tax_digits'), ('afip', '0004_auto_20150916_1934'), ('afip', '0005_auto_20150918_0115'), ('afip', '0006_auto_20151212_1431'), ('afip', '0007_auto_20151212_1754'), ('afip', '0008_auto_20151212_1820'), ('afip', '0009_auto_20151214_1836'), ('afip', '0010_receiptpdf_pdffile_uploadto'), ('afip', '0011_receipt_entry_vat'), ('afip', '0012_taxpayer_profile_one_to_one'), ('afip', '0013_taxpayer_is_sandboxed'), ('afip', '0014_no_partially_validated_receiptvalidations'), ('afip', '0015_receiptentry__amount_to_quantity'), ('afip', '0016_auto_20170529_2012'), ('afip', '0017_receipt_issued_date'), ('afip', '0018_on_delete_fks'), ('afip', '0019_receiptvalidation__processed_date'), ('afip', '0020_backfill_receiptvalidation__processed_date'), ('afip', '0021_drop_batches'), ('afip', '0022_auto_misc_tweaks'), ('afip', '0023_taxpayer__certs_blank'), ('afip', '0024_taxpayer__certificate_expiration'), ('afip', '0025_receipt__default_currency'), ('afip', '0026_vat_conditions'), ('afip', '0027_taxpayer__active_since'), ('afip', '0028_taxpayer__copy_active_since'), ('afip', '0029_drop__taxpayerprofile__active_since'), ('afip', '0030_receiptpdf_client_vat_condition'), ('afip', '0031_receiptpdf__copy_vat_condition'), ('afip', '0032_receiptpdf__client_vat__notnull'), ('afip', '0033_receiptpdf__pdf_file__help_text'), ('afip', '0034_vat_condition_choices'), ('afip', '0035_receiptentry__vat__blankable'), ('afip', '0036_receiptpdf__client_address__blank')]

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='AuthTicket',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('unique_id', models.IntegerField(default=django_afip.models.AuthTicket.default_unique_id, verbose_name='unique id')),
                ('generated', models.DateTimeField(default=django_afip.models.AuthTicket.default_generated, verbose_name='generated')),
                ('expires', models.DateTimeField(default=django_afip.models.AuthTicket.default_expires, verbose_name='expires')),
                ('service', models.CharField(help_text='Service for which this ticket has been authorized', max_length=6, verbose_name='service')),
                ('token', models.TextField(verbose_name='token')),
                ('signature', models.TextField(verbose_name='signature')),
            ],
            options={
                'verbose_name': 'authorization ticket',
                'verbose_name_plural': 'authorization tickets',
            },
        ),
        migrations.CreateModel(
            name='ConceptType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=3, verbose_name='code')),
                ('description', models.CharField(max_length=250, verbose_name='description')),
                ('valid_from', models.DateField(blank=True, null=True, verbose_name='valid from')),
                ('valid_to', models.DateField(blank=True, null=True, verbose_name='valid until')),
            ],
            options={
                'verbose_name': 'concept type',
                'verbose_name_plural': 'concept types',
            },
        ),
        migrations.CreateModel(
            name='CurrencyType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=3, verbose_name='code')),
                ('description', models.CharField(max_length=250, verbose_name='description')),
                ('valid_from', models.DateField(blank=True, null=True, verbose_name='valid from')),
                ('valid_to', models.DateField(blank=True, null=True, verbose_name='valid until')),
            ],
            options={
                'verbose_name': 'currency type',
                'verbose_name_plural': 'currency types',
            },
        ),
        migrations.CreateModel(
            name='DocumentType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=3, verbose_name='code')),
                ('description', models.CharField(max_length=250, verbose_name='description')),
                ('valid_from', models.DateField(blank=True, null=True, verbose_name='valid from')),
                ('valid_to', models.DateField(blank=True, null=True, verbose_name='valid until')),
            ],
            options={
                'verbose_name': 'document type',
                'verbose_name_plural': 'document types',
            },
        ),
        migrations.CreateModel(
            name='Observation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.PositiveSmallIntegerField(verbose_name='code')),
                ('message', models.CharField(max_length=255, verbose_name='message')),
            ],
            options={
                'verbose_name': 'observation',
                'verbose_name_plural': 'observations',
            },
        ),
        migrations.CreateModel(
            name='PointOfSales',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('number', models.PositiveSmallIntegerField(verbose_name='number')),
                ('issuance_type', models.CharField(help_text='Indicates if thie POS emits using CAE and CAEA.', max_length=8, verbose_name='issuance type')),
                ('blocked', models.BooleanField(verbose_name='blocked')),
                ('drop_date', models.DateField(blank=True, null=True, verbose_name='drop date')),
            ],
            options={
                'verbose_name': 'point of sales',
                'verbose_name_plural': 'points of sales',
            },
        ),
        migrations.CreateModel(
            name='Receipt',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('document_number', models.BigIntegerField(help_text='The document number of the customer to whom this receipt is addressed', verbose_name='document number')),
                ('receipt_number', models.PositiveIntegerField(blank=True, help_text='If left blank, the next valid number will assigned when validating the receipt.', null=True, verbose_name='receipt number')),
                ('issued_date', models.DateField(help_text='Can diverge up to 5 days for good, or 10 days otherwise', verbose_name='issued date')),
                ('total_amount', models.DecimalField(decimal_places=2, help_text='Must be equal to the sum of net_taxed, exempt_amount, net_taxes, and all taxes and vats.', max_digits=15, verbose_name='total amount')),
                ('net_untaxed', models.DecimalField(decimal_places=2, help_text='The total amount to which taxes do not apply. For C-type receipts, this must be zero.', max_digits=15, verbose_name='total untaxable amount')),
                ('net_taxed', models.DecimalField(decimal_places=2, help_text='The total amount to which taxes apply. For C-type receipts, this is equal to the subtotal.', max_digits=15, verbose_name='total taxable amount')),
                ('exempt_amount', models.DecimalField(decimal_places=2, help_text='Only for categories which are tax-exempt. For C-type receipts, this must be zero.', max_digits=15, verbose_name='exempt amount')),
                ('service_start', models.DateField(blank=True, help_text='Date on which a service started. No applicable for goods.', null=True, verbose_name='service start date')),
                ('service_end', models.DateField(blank=True, help_text='Date on which a service ended. No applicable for goods.', null=True, verbose_name='service end date')),
                ('expiration_date', models.DateField(blank=True, help_text='Date on which this receipt expires. No applicable for goods.', null=True, verbose_name='receipt expiration date')),
                ('currency_quote', models.DecimalField(decimal_places=6, default=1, help_text='Quote of the day for the currency used in the receipt', max_digits=10, verbose_name='currency quote')),
                ('concept', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='receipts', to='afip.ConceptType', verbose_name='concept')),
                ('currency', models.ForeignKey(default=django_afip.models.first_currency, help_text='Currency in which this receipt is issued.', on_delete=django.db.models.deletion.PROTECT, related_name='documents', to='afip.CurrencyType', verbose_name='currency')),
                ('document_type', models.ForeignKey(help_text='The document type of the customer to whom this receipt is addressed', on_delete=django.db.models.deletion.PROTECT, related_name='receipts', to='afip.DocumentType', verbose_name='document type')),
                ('point_of_sales', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='receipts', to='afip.PointOfSales', verbose_name='point of sales')),
            ],
            options={
                'verbose_name': 'receipt',
                'verbose_name_plural': 'receipts',
                'ordering': ('issued_date',),
            },
        ),
        migrations.CreateModel(
            name='ReceiptEntry',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(max_length=128, verbose_name='description')),
                ('quantity', models.PositiveSmallIntegerField(verbose_name='quantity')),
                ('unit_price', models.DecimalField(decimal_places=2, help_text='Price per unit before vat or taxes.', max_digits=15, verbose_name='unit price')),
                ('receipt', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='entries', to='afip.Receipt', verbose_name='receipt')),
            ],
            options={
                'verbose_name': 'receipt entry',
                'verbose_name_plural': 'receipt entries',
            },
        ),
        migrations.CreateModel(
            name='ReceiptPDF',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('pdf_file', models.FileField(blank=True, help_text='The actual file which contains the PDF data.', null=True, upload_to='receipts', verbose_name='pdf file')),
                ('issuing_name', models.CharField(max_length=128, verbose_name='issuing name')),
                ('issuing_address', models.TextField(verbose_name='issuing address')),
                ('issuing_email', models.CharField(blank=True, max_length=128, null=True, verbose_name='issuing email')),
                ('vat_condition', models.CharField(choices=[('IVA Responsable Inscripto', 'IVA Responsable Inscripto'), ('IVA Responsable No Inscripto', 'IVA Responsable No Inscripto'), ('IVA Exento', 'IVA Exento'), ('No Responsable IVA', 'No Responsable IVA'), ('Responsable Monotributo', 'Responsable Monotributo')], max_length=48, verbose_name='vat condition')),
                ('gross_income_condition', models.CharField(max_length=48, verbose_name='gross income condition')),
                ('client_name', models.CharField(max_length=128, verbose_name='client name')),
                ('client_address', models.TextField(blank=True, verbose_name='client address')),
                ('client_vat_condition', models.CharField(choices=[('IVA Responsable Inscripto', 'IVA Responsable Inscripto'), ('IVA Responsable No Inscripto', 'IVA Responsable No Inscripto'), ('IVA Sujeto Exento', 'IVA Sujeto Exento'), ('Consumidor Final', 'Consumidor Final'), ('Responsable Monotributo', 'Responsable Monotributo'), ('Proveedor del Exterior', 'Proveedor del Exterior'), ('Cliente del Exterior', 'Cliente del Exterior'), ('IVA Liberado - Ley Nº 19.640', 'IVA Liberado - Ley Nº 19.640'), ('IVA Responsable Inscripto - Agente de Percepción', 'IVA Responsable Inscripto - Agente de Percepción'), ('Monotributista Social', 'Monotributista Social'), ('IVA no alcanzado', 'IVA no alcanzado')], max_length=48, verbose_name='client vat condition')),
                ('sales_terms', models.CharField(help_text='Should be something like "Cash", "Payable in 30 days", etc', max_length=48, verbose_name='sales terms')),
                ('receipt', models.OneToOneField(on_delete=django.db.models.deletion.PROTECT, to='afip.Receipt', verbose_name='receipt')),
            ],
            options={
                'verbose_name': 'receipt pdf',
                'verbose_name_plural': 'receipt pdfs',
            },
        ),
        migrations.CreateModel(
            name='ReceiptType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=3, verbose_name='code')),
                ('description', models.CharField(max_length=250, verbose_name='description')),
                ('valid_from', models.DateField(blank=True, null=True, verbose_name='valid from')),
                ('valid_to', models.DateField(blank=True, null=True, verbose_name='valid until')),
            ],
            options={
                'verbose_name': 'receipt type',
                'verbose_name_plural': 'receipt types',
            },
        ),
        migrations.CreateModel(
            name='ReceiptValidation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('result', models.CharField(choices=[('A', 'approved'), ('R', 'rejected')], help_text='Indicates whether the validation was succesful or not', max_length=1, verbose_name='result')),
                ('processed_date', models.DateTimeField(verbose_name='processed date')),
                ('cae', models.CharField(help_text='The CAE as returned by the AFIP', max_length=14, verbose_name='cae')),
                ('cae_expiration', models.DateField(help_text='The CAE expiration as returned by the AFIP', verbose_name='cae expiration')),
                ('observations', models.ManyToManyField(help_text='The observations as returned by the AFIP. These are generally present for failed validations.', related_name='validations', to='afip.Observation', verbose_name='observations')),
                ('receipt', models.OneToOneField(help_text='The Receipt for which this validation applies', on_delete=django.db.models.deletion.PROTECT, related_name='validation', to='afip.Receipt', verbose_name='receipt')),
            ],
            options={
                'verbose_name': 'receipt validation',
                'verbose_name_plural': 'receipt validations',
            },
        ),
        migrations.CreateModel(
            name='Tax',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(max_length=80, verbose_name='description')),
                ('base_amount', models.DecimalField(decimal_places=2, max_digits=15, verbose_name='base amount')),
                ('aliquot', models.DecimalField(decimal_places=2, max_digits=5, verbose_name='aliquot')),
                ('amount', models.DecimalField(decimal_places=2, max_digits=15, verbose_name='amount')),
                ('receipt', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='taxes', to='afip.Receipt')),
            ],
            options={
                'verbose_name': 'tax',
                'verbose_name_plural': 'taxes',
            },
        ),
        migrations.CreateModel(
            name='TaxPayer',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(help_text='A friendly name to recognize this taxpayer.', max_length=32, verbose_name='name')),
                ('key', models.FileField(blank=True, null=True, upload_to='', verbose_name='key')),
                ('certificate', models.FileField(blank=True, null=True, upload_to='', verbose_name='certificate')),
                ('cuit', models.BigIntegerField(verbose_name='cuit')),
                ('is_sandboxed', models.BooleanField(help_text='Indicates if this taxpayer interacts with the sandbox servers rather than the production servers', verbose_name='is sandboxed')),
                ('certificate_expiration', models.DateTimeField(editable=False, help_text='Stores expiration for the current certificate. Note that this field is updated pre-save, so the value may be invalid for unsaved models.', null=True, verbose_name='certificate expiration')),
                ('active_since', models.DateField(help_text='Date since which this taxpayer has been legally active.', verbose_name='active since')),
            ],
            options={
                'verbose_name': 'taxpayer',
                'verbose_name_plural': 'taxpayers',
            },
        ),
        migrations.CreateModel(
            name='TaxPayerProfile',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('issuing_name', models.CharField(max_length=128, verbose_name='issuing name')),
                ('issuing_address', models.TextField(verbose_name='issuing address')),
                ('issuing_email', models.CharField(blank=True, max_length=128, null=True, verbose_name='issuing email')),
                ('vat_condition', models.CharField(choices=[('IVA Responsable Inscripto', 'IVA Responsable Inscripto'), ('IVA Responsable No Inscripto', 'IVA Responsable No Inscripto'), ('IVA Exento', 'IVA Exento'), ('No Responsable IVA', 'No Responsable IVA'), ('Responsable Monotributo', 'Responsable Monotributo')], max_length=48, verbose_name='vat condition')),
                ('gross_income_condition', models.CharField(max_length=48, verbose_name='gross income condition')),
                ('sales_terms', models.CharField(help_text='The terms of the sale printed onto receipts by default (eg: single payment, checking account, etc).', max_length=48, verbose_name='sales terms')),
                ('taxpayer', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='profile', to='afip.TaxPayer', verbose_name='taxpayer')),
            ],
            options={
                'verbose_name': 'taxpayer profile',
                'verbose_name_plural': 'taxpayer profiles',
            },
        ),
        migrations.CreateModel(
            name='TaxType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=3, verbose_name='code')),
                ('description', models.CharField(max_length=250, verbose_name='description')),
                ('valid_from', models.DateField(blank=True, null=True, verbose_name='valid from')),
                ('valid_to', models.DateField(blank=True, null=True, verbose_name='valid until')),
            ],
            options={
                'verbose_name': 'tax type',
                'verbose_name_plural': 'tax types',
            },
        ),
        migrations.CreateModel(
            name='Vat',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('base_amount', models.DecimalField(decimal_places=2, max_digits=15, verbose_name='base amount')),
                ('amount', models.DecimalField(decimal_places=2, max_digits=15, verbose_name='amount')),
                ('receipt', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='vat', to='afip.Receipt')),
            ],
            options={
                'verbose_name': 'vat',
                'verbose_name_plural': 'vat',
            },
        ),
        migrations.CreateModel(
            name='VatType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('code', models.CharField(max_length=3, verbose_name='code')),
                ('description', models.CharField(max_length=250, verbose_name='description')),
                ('valid_from', models.DateField(blank=True, null=True, verbose_name='valid from')),
                ('valid_to', models.DateField(blank=True, null=True, verbose_name='valid until')),
            ],
            options={
                'verbose_name': 'vat type',
                'verbose_name_plural': 'vat types',
            },
        ),
        migrations.AddField(
            model_name='vat',
            name='vat_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='afip.VatType', verbose_name='vat type'),
        ),
        migrations.AddField(
            model_name='tax',
            name='tax_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='afip.TaxType', verbose_name='tax type'),
        ),
        migrations.AddField(
            model_name='receiptentry',
            name='vat',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.PROTECT, related_name='receipt_entries', to='afip.VatType', verbose_name='vat'),
        ),
        migrations.AddField(
            model_name='receipt',
            name='receipt_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, related_name='receipts', to='afip.ReceiptType', verbose_name='receipt type'),
        ),
        migrations.AddField(
            model_name='receipt',
            name='related_receipts',
            field=models.ManyToManyField(blank=True, to='afip.Receipt', verbose_name='related receipts'),
        ),
        migrations.AddField(
            model_name='pointofsales',
            name='owner',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='points_of_sales', to='afip.TaxPayer', verbose_name='owner'),
        ),
        migrations.AddField(
            model_name='authticket',
            name='owner',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='auth_tickets', to='afip.TaxPayer', verbose_name='owner'),
        ),
        migrations.AlterUniqueTogether(
            name='receipt',
            unique_together=set([('point_of_sales', 'receipt_type', 'receipt_number')]),
        ),
        migrations.AlterUniqueTogether(
            name='pointofsales',
            unique_together=set([('number', 'owner')]),
        ),
    ]



"""Tests for provided template tags."""
from django.test import TestCase

from django_afip.templatetags.django_afip import format_cuit


class FormatCuitTagTestCase(TestCase):
    """Test the format_cuit tag."""

    def test_good_string_input(self):
        """Test valid string inputs."""
        self.assertEqual(
            format_cuit('20329642330'),
            '20-32964233-0'
        )
        self.assertEqual(
            format_cuit('20-32964233-0'),
            '20-32964233-0'
        )

    def test_good_numeric_input(self):
        """Test valid numerical input."""
        self.assertEqual(
            format_cuit(20329642330),
            '20-32964233-0'
        )

    def test_bad_string_input(self):
        """Test invalid string input."""
        self.assertEqual(
            format_cuit('blah blah'),
            'blah blah'
        )

    def test_bad_numeric_input(self):
        """Test invalid numerical input."""
        self.assertEqual(
            format_cuit(1234),
            1234
        )



import logging
from io import BytesIO

from barcode import ITF
from barcode.writer import ImageWriter
from django.utils.functional import cached_property


logger = logging.getLogger(__name__)


class ImageWitoutTextWriter(ImageWriter):
    def _paint_text(self, xpos, ypos):
        pass


class ReceiptBarcodeGenerator:
    def __init__(self, receipt):
        self._receipt = receipt

    @cached_property
    def numbers(self):
        """"
        Returns the barcode's number without the verification digit.

        :return: list(int)
        """
        numstring = '{:011d}{:02d}{:04d}{}{}'.format(
            self._receipt.point_of_sales.owner.cuit,  # 11 digits
            int(self._receipt.receipt_type.code),  # 2 digits
            self._receipt.point_of_sales.number,  # point of sales
            self._receipt.validation.cae,  # 14 digits
            self._receipt.validation.cae_expiration.strftime('%Y%m%d'),  # 8
        )
        return [int(num) for num in numstring]

    @staticmethod
    def verification_digit(numbers):
        """
        Returns the verification digit for a given numbre.

        The verification digit is calculated as follows:

        * A = sum of all even-positioned numbers
        * B = A * 3
        * C = sum of all odd-positioned numbers
        * D = B + C
        * The results is the smallset number N, such that (D + N) % 10 == 0

        NOTE: Afip's documentation seems to have odd an even mixed up in the
        explanation, but all examples follow the above algorithm.

        :param list(int) numbers): The numbers for which the digits is to be
            calculated.
        :return: int
        """
        a = sum(numbers[::2])
        b = a * 3
        c = sum(numbers[1::2])
        d = b + c
        e = d % 10
        if e == 0:
            return e
        return 10 - e

    @cached_property
    def full_number(self):
        """
        Returns the full number including the verification digit.

        :return: str
        """
        return '{}{}'.format(
            ''.join(str(n) for n in self.numbers),
            ReceiptBarcodeGenerator.verification_digit(self.numbers),
        )

    def generate_barcode(self, writer_class=ImageWitoutTextWriter):
        rv = BytesIO()
        ITF(self.full_number, writer=writer_class()).write(rv)

        return rv.getvalue()



from datetime import datetime

from django.test import TestCase
from factory.django import FileField
from freezegun import freeze_time
from OpenSSL import crypto

from testapp.testmain import fixtures


class TestTaxPayerKeyManagement(TestCase):

    def test_key_generation(self):
        taxpayer = fixtures.TaxPayerFactory(key=None)
        taxpayer.generate_key()

        key = taxpayer.key.file.read().decode()
        self.assertEqual(key.splitlines()[0], '-----BEGIN PRIVATE KEY-----')
        self.assertEqual(key.splitlines()[-1], '-----END PRIVATE KEY-----')

        loaded_key = crypto.load_privatekey(crypto.FILETYPE_PEM, key)
        self.assertIsInstance(loaded_key, crypto.PKey)

    def test_dont_overwrite_keys(self):
        text = "Hello! I'm not really a key :D".encode()
        taxpayer = fixtures.TaxPayerFactory(key=FileField(data=text))

        taxpayer.generate_key()
        key = taxpayer.key.read()

        self.assertEqual(text, key)

    def test_overwrite_keys_force(self):
        text = "Hello! I'm not really a key :D".encode()
        taxpayer = fixtures.TaxPayerFactory(key__data=text)

        taxpayer.generate_key(force=True)
        key = taxpayer.key.file.read().decode()

        self.assertNotEqual(text, key)
        self.assertEqual(key.splitlines()[0], '-----BEGIN PRIVATE KEY-----')
        self.assertEqual(key.splitlines()[-1], '-----END PRIVATE KEY-----')

        loaded_key = crypto.load_privatekey(crypto.FILETYPE_PEM, key)
        self.assertIsInstance(loaded_key, crypto.PKey)

    @freeze_time(datetime.fromtimestamp(1489537017))
    def test_csr_generation(self):
        taxpayer = fixtures.TaxPayerFactory(key=None)
        taxpayer.generate_key()

        csr_file = taxpayer.generate_csr()
        csr = csr_file.read().decode()

        self.assertEqual(
            csr.splitlines()[0],
            '-----BEGIN CERTIFICATE REQUEST-----'
        )

        self.assertEqual(
            csr.splitlines()[-1],
            '-----END CERTIFICATE REQUEST-----'
        )

        loaded_csr = crypto.load_certificate_request(crypto.FILETYPE_PEM, csr)
        self.assertIsInstance(loaded_csr, crypto.X509Req)

        expected_components = [
            (b'O', b'John Smith'),
            (b'CN', b'djangoafip1489537017'),
            (b'serialNumber', b'CUIT 20329642330'),
        ]

        self.assertEqual(
            expected_components,
            loaded_csr.get_subject().get_components()
        )


class TaxPayerCertificateObjectTestCase(TestCase):
    def test_certificate_object(self):
        taxpayer = fixtures.TaxPayerFactory()
        cert = taxpayer.certificate_object

        self.assertIsInstance(cert, crypto.X509)

    def test_null_certificate_object(self):
        taxpayer = fixtures.TaxPayerFactory(certificate=None)
        cert = taxpayer.certificate_object

        self.assertIsNone(cert)


class TaxPayerExpirationTestCase(TestCase):
    def test_expiration_getter(self):
        taxpayer = fixtures.TaxPayerFactory()
        expiration = taxpayer.get_certificate_expiration()

        self.assertIsInstance(expiration, datetime)

    def test_expiration_signal_update(self):
        taxpayer = fixtures.TaxPayerFactory(certificate_expiration=None)
        taxpayer.save()
        expiration = taxpayer.certificate_expiration

        self.assertIsInstance(expiration, datetime)



from datetime import date, datetime

import pytz
from django.test import TestCase

from django_afip import parsers


class ParseDatetimeTestCase(TestCase):
    def test_parse_null(self):
        self.assertEqual(parsers.parse_datetime('NULL'), None)

    def test_parse_none(self):
        self.assertEqual(parsers.parse_datetime(None), None)

    def test_parse_datetimes(self):
        tz = pytz.timezone(pytz.country_timezones['ar'][0])
        self.assertEqual(
            parsers.parse_datetime('20170730154330'),
            datetime(2017, 7, 30, 15, 43, 30, tzinfo=tz),
        )


class ParseDateTestCase(TestCase):
    def test_parse_null(self):
        self.assertEqual(parsers.parse_date('NULL'), None)

    def test_parse_none(self):
        self.assertEqual(parsers.parse_date(None), None)

    def test_parse_dates(self):
        self.assertEqual(
            parsers.parse_date('20170730'),
            date(2017, 7, 30),
        )


class ParseString(TestCase):
    def test_weirdly_encoded(self):
        # This is the encoding AFIP sometimes uses:
        string = 'AÃ±adir paÃ\xads'
        self.assertEqual(parsers.parse_string(string), 'Añadir país')



from unittest.mock import patch

from django.test import TestCase

from django_afip.clients import get_client


class TestGetService(TestCase):
    def test_services_are_cached(self):
        service1 = get_client('wsfe', False)
        with patch.dict('django_afip.clients.wsdls', values={}, clear=True):
            service2 = get_client('wsfe', False)

        self.assertEqual(service1, service2)

    def test_inexisting_service(self):
        with self.assertRaisesMessage(
            ValueError,
            'Unknown service name, nonexistant'
        ):
            get_client('nonexistant', False)



from datetime import datetime

import pytz

TZ_AR = pytz.timezone(pytz.country_timezones['ar'][0])


def parse_datetime(datestring):
    if datestring == 'NULL' or datestring is None:
        return None
    return datetime.strptime(datestring, '%Y%m%d%H%M%S') \
        .replace(tzinfo=TZ_AR)


def parse_date(datestring):
    if datestring == 'NULL' or datestring is None:
        return None
    return datetime.strptime(datestring, '%Y%m%d').date()


def parse_string(string):
    """Re-encodes strings from AFIP's weird encoding to UTF-8."""
    try:
        return string.encode('latin-1').decode()
    except UnicodeDecodeError:
        # It looks like SOME errors are plain UTF-8 text.
        return string



"""testapp URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/1.8/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')
Including another URLconf
    1. Add an import:  from blog import urls as blog_urls
    2. Add a URL to urlpatterns:  url(r'^blog/', include(blog_urls))
"""
from django.conf import settings
from django.conf.urls import url
from django.contrib import admin
from django.views.static import serve

from django_afip import views

urlpatterns = [
    url(r'^admin/', admin.site.urls),
    url(
        r'^invoices/pdf/(?P<pk>\d+)$',
        views.ReceiptPDFView.as_view(),
        name='receipt_displaypdf_view',
    ),
    url(
        r'^invoices/pdf/(?P<pk>\d+)$',
        views.ReceiptPDFDownloadView.as_view(),
        name='receipt_pdf_view',
    ),
    url(
        r'^media/(?P<path>.*)$',
        serve,
        {'document_root': settings.MEDIA_ROOT},
    ),
]



"""
Automatically list fields from django models.

Based on https://djangosnippets.org/snippets/2533/
"""

import inspect

from django.db import models
from django.utils.encoding import force_text
from django.utils.html import strip_tags


def process_docstring(app, what, name, obj, options, lines):
    """
    Process docstrings for django models.

    Process docstrings for django models and add field description from
    their help_text attribute.
    """
    # Only look at objects that inherit from Django's base model class
    if inspect.isclass(obj) and issubclass(obj, models.Model):
        # Grab the field list from the meta class
        fields = obj._meta.get_fields()

        for field in fields:
            if field.name == 'id':
                continue

            if not hasattr(field, 'help_text') and \
               not hasattr(field, 'verbose'):
                # XXX: log these?
                continue

            help_text = strip_tags(force_text(field.help_text))
            verbose_name = force_text(field.verbose_name).capitalize()

            if help_text:
                # Add the model field to the end of the docstring as a param
                # using the help text as the description
                lines.append(':param %s: %s' % (field.attname, help_text))
            else:
                # Add the model field to the end of the docstring as a param
                # using the verbose name as the description
                lines.append(':param %s: %s' % (field.attname, verbose_name))

            # Add the field's type to the docstring
            if isinstance(field, models.ForeignKey):
                to = field.remote_field.model
                lines.append(':type %s: %s to :class:`~%s.%s`' % (
                    field.attname,
                    type(field).__name__,
                    to.__module__,
                    to.__name__
                ))
            else:
                lines.append(':type %s: %s' % (
                    field.attname, type(field).__name__)
                )

    # Return the extended docstring
    return lines


def setup(app):
    """Register the docstring processor with sphinx."""
    app.connect('autodoc-process-docstring', process_docstring)



from unittest.mock import call, MagicMock, patch

from django.test import TestCase

from django_afip import exceptions, models
from testapp.testmain import fixtures
from testapp.testmain.tests.testcases import PopulatedLiveAfipTestCase


class ReceiptQuerySetTestCase(TestCase):

    def test_default_manager(self):
        self.assertIsInstance(
            models.Receipt.objects.all(),
            models.ReceiptQuerySet,
        )

    def test_validate(self):
        receipt = fixtures.ReceiptFactory()
        queryset = models.Receipt.objects.filter(pk=receipt.pk)
        ticket = MagicMock()

        with patch(
            'django_afip.models.ReceiptQuerySet._assign_numbers', spec=True,
        ) as mocked_assign_numbers, patch(
            'django_afip.models.ReceiptQuerySet._validate', spec=True,
        ) as mocked__validate:
            queryset.validate(ticket)

        self.assertEqual(mocked_assign_numbers.call_count, 1)

        self.assertEqual(mocked__validate.call_count, 1)
        self.assertEqual(mocked__validate.call_args, call(ticket))

        # TODO: Also another tests that checks that we only pass filtered-out
        # receipts


class ReceiptManagerTestCase(TestCase):

    def test_default_manager(self):
        self.assertIsInstance(models.Receipt.objects, models.ReceiptManager)


class ReceiptTestCase(TestCase):

    def test_validate(self):
        receipt = fixtures.ReceiptFactory()
        ticket = MagicMock()
        self.called = False

        def fake_validate(qs, ticket=None):
            self.assertQuerysetEqual(qs, [receipt.pk], lambda r: r.pk)
            self.called = True

        with patch(
            'django_afip.models.ReceiptQuerySet.validate',
            fake_validate,
        ):
            receipt.validate(ticket)

        self.assertTrue(self.called)


class ReceiptSuccessfulValidateTestCase(PopulatedLiveAfipTestCase):

    def setUp(self):
        super().setUp()
        self.receipt = fixtures.ReceiptFactory(
            point_of_sales=models.PointOfSales.objects.first(),
        )
        fixtures.VatFactory(vat_type__code=5, receipt=self.receipt)
        fixtures.TaxFactory(tax_type__code=3, receipt=self.receipt)

    def test_validation(self):
        """Test validating valid receipts."""
        errs = self.receipt.validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            self.receipt.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 1)


class ReceiptFailedValidateTestCase(PopulatedLiveAfipTestCase):

    def setUp(self):
        super().setUp()
        self.receipt = fixtures.ReceiptFactory(
            document_type__code=80,
            point_of_sales=models.PointOfSales.objects.first(),
        )
        fixtures.VatFactory(vat_type__code=5, receipt=self.receipt)
        fixtures.TaxFactory(tax_type__code=3, receipt=self.receipt)

    def test_failed_validation(self):
        """Test validating valid receipts."""
        errs = self.receipt.validate()

        self.assertEqual(len(errs), 1)
        # FIXME: We're not creating rejection entries
        # self.assertEqual(len(errs), 1)
        # self.assertEqual(
        #     receipt.validation.result,
        #     models.ReceiptValidation.RESULT_REJECTED,
        # )
        self.assertEqual(models.ReceiptValidation.objects.count(), 0)

    def test_raise_validation(self):
        """Test validating valid receipts."""

        with self.assertRaisesRegex(
            exceptions.ValidationError,
            # Note: AFIP apparently edited this message and added a typo:
            'DocNro 203012345 no se encuentra registrado en los padrones',
        ):
            self.receipt.validate(raise_=True)

        # FIXME: We're not creating rejection entries
        # self.assertEqual(
        #     receipt.validation.result,
        #     models.ReceiptValidation.RESULT_REJECTED,
        # )
        self.assertEqual(models.ReceiptValidation.objects.count(), 0)


class ReceiptIsValidatedTestCase(TestCase):
    def test_not_validated(self):
        receipt = fixtures.ReceiptFactory()
        self.assertEqual(receipt.is_validated, False)

    def test_validated(self):
        receipt = fixtures.ReceiptFactory(receipt_number=1)
        fixtures.ReceiptValidationFactory(receipt=receipt)
        self.assertEqual(receipt.is_validated, True)

    def test_failed_validation(self):
        # These should never really exist,but oh well:
        receipt = fixtures.ReceiptFactory()
        fixtures.ReceiptValidationFactory(
            receipt=receipt,
            result=models.ReceiptValidation.RESULT_REJECTED,
        )
        self.assertEqual(receipt.is_validated, False)

        receipt = fixtures.ReceiptFactory(receipt_number=1)
        fixtures.ReceiptValidationFactory(
            receipt=receipt,
            result=models.ReceiptValidation.RESULT_REJECTED,
        )
        self.assertEqual(receipt.is_validated, False)


class ReceiptDefaultCurrencyTestCase(TestCase):
    def test_no_currencies(self):
        receipt = models.Receipt()
        with self.assertRaises(models.CurrencyType.DoesNotExist):
            receipt.currency

    def test_multieple_currencies(self):
        c1 = fixtures.CurrencyTypeFactory(pk=2)
        c2 = fixtures.CurrencyTypeFactory(pk=1)
        c3 = fixtures.CurrencyTypeFactory(pk=3)

        receipt = models.Receipt()
        self.assertNotEqual(receipt.currency, c1)
        self.assertEqual(receipt.currency, c2)
        self.assertNotEqual(receipt.currency, c3)


class ReceiptTotalVatTestCase(TestCase):
    def test_no_vat(self):
        receipt = fixtures.ReceiptFactory()

        self.assertEqual(receipt.total_vat, 0)

    def test_multiple_vats(self):
        receipt = fixtures.ReceiptFactory()
        fixtures.VatFactory(receipt=receipt)
        fixtures.VatFactory(receipt=receipt)

        self.assertEqual(receipt.total_vat, 42)

    def test_proper_filtering(self):
        receipt = fixtures.ReceiptFactory()
        fixtures.VatFactory(receipt=receipt)
        fixtures.VatFactory()

        self.assertEqual(receipt.total_vat, 21)


class ReceiptTotalTaxTestCase(TestCase):
    def test_no_tax(self):
        receipt = fixtures.ReceiptFactory()

        self.assertEqual(receipt.total_tax, 0)

    def test_multiple_taxes(self):
        receipt = fixtures.ReceiptFactory()
        fixtures.TaxFactory(receipt=receipt)
        fixtures.TaxFactory(receipt=receipt)

        self.assertEqual(receipt.total_tax, 18)

    def test_proper_filtering(self):
        receipt = fixtures.ReceiptFactory()
        fixtures.TaxFactory(receipt=receipt)
        fixtures.TaxFactory()

        self.assertEqual(receipt.total_tax, 9)


class CurrencyTypeStrTestCase(TestCase):
    def test_success(self):
        currency_type = models.CurrencyType(
            code='011',
            description='Pesos Uruguayos',
        )
        self.assertEqual(str(currency_type), 'Pesos Uruguayos (011)')



"""Tests for AFIP-WS related classes."""
import os
from datetime import datetime, timedelta
from unittest.mock import patch

from django.conf import settings
from django.core import management
from django.test import tag, TestCase
from django.utils.timezone import now
from factory.django import FileField

from django_afip import exceptions, models
from testapp.testmain import fixtures
from testapp.testmain.tests.testcases import (
    LiveAfipTestCase,
    PopulatedLiveAfipTestCase,
)


@tag('live')
class AuthTicketTest(TestCase):
    """Test AuthTicket methods."""

    def test_bad_cuit(self):
        """Test using the wrong cuit for a key pair."""

        taxpayer = fixtures.TaxPayerFactory(cuit=20329642339)
        taxpayer.create_ticket('wsfe')

        with self.assertRaisesRegex(
            exceptions.AfipException,
            # Note: AFIP apparently edited this message and added a typo:
            'ValidacionDeToken: No apareci[oó] CUIT en lista de relaciones:',
        ):
            models.populate_all()

    def test_bogus_certificate_exception(self):
        """Test that using a junk ceritificates raises as expected."""

        # New TaxPayers will fail to save with an invalid cert, but many
        # systems may have very old TaxPayers, externally created, or other
        # stuff, so this scenario might still be possible.
        with patch(
            'django_afip.models.TaxPayer.get_certificate_expiration',
            spec=True,
            return_value=None,
        ):
            taxpayer = fixtures.TaxPayerFactory(
                key=FileField(data=b'Blah'),
                certificate=FileField(data=b'Blah'),
            )

        with self.assertRaises(exceptions.CorruptCertificate) as e:
            taxpayer.create_ticket('wsfe')

        self.assertNotIsInstance(e, exceptions.AfipException)

    def test_no_active_taxpayer(self):
        """Test that no TaxPayers raises an understandable error."""
        with self.assertRaisesMessage(
            exceptions.AuthenticationError,
            'There are no taxpayers to generate a ticket.',
        ):
            models.AuthTicket.objects.get_any_active('wsfe')

    def test_expired_certificate_exception(self):
        """Test that using an expired ceritificate raises as expected."""
        with open(
            os.path.join(settings.BASE_DIR, 'test_expired.key'),
        ) as key, open(
            os.path.join(settings.BASE_DIR, 'test_expired.crt'),
        ) as crt:
            taxpayer = fixtures.TaxPayerFactory(
                key=FileField(from_file=key),
                certificate=FileField(from_file=crt),
            )

        with self.assertRaises(exceptions.CertificateExpired):
            taxpayer.create_ticket('wsfe')

    def test_untrusted_certificate_exception(self):
        """
        Test that using an untrusted ceritificate raises as expected.
        """
        # Note that we hit production with a sandbox cert here:
        taxpayer = fixtures.TaxPayerFactory(is_sandboxed=False)

        with self.assertRaises(exceptions.UntrustedCertificate):
            taxpayer.create_ticket('wsfe')


class PopulationTest(LiveAfipTestCase):
    """
    Tests models population view.

    As a side effect, also test valid ticket generation.
    """

    def test_population_command(self):
        """Test the afipmetadata command."""
        management.call_command('afipmetadata')

        receipts = models.ReceiptType.objects.count()
        concepts = models.ConceptType.objects.count()
        documents = models.DocumentType.objects.count()
        vat = models.VatType.objects.count()
        tax = models.TaxType.objects.count()
        currencies = models.CurrencyType.objects.count()

        self.assertGreater(receipts, 0)
        self.assertGreater(concepts, 0)
        self.assertGreater(documents, 0)
        self.assertGreater(vat, 0)
        self.assertGreater(tax, 0)
        self.assertGreater(currencies, 0)


class TaxPayerTest(LiveAfipTestCase):
    """Test TaxPayer methods."""

    def test_fetch_points_of_sale(self):
        """Test the ``fetch_points_of_sales`` method."""
        taxpayer = models.TaxPayer.objects.first()
        taxpayer.fetch_points_of_sales()

        points_of_sales = models.PointOfSales.objects.count()
        self.assertGreater(points_of_sales, 0)


class ReceiptQuerySetTestCase(PopulatedLiveAfipTestCase):
    """Test ReceiptQuerySet methods."""

    def _good_receipt(self):
        receipt = fixtures.ReceiptFactory(
            point_of_sales=models.PointOfSales.objects.first(),
        )
        fixtures.VatFactory(vat_type__code=5, receipt=receipt)
        fixtures.TaxFactory(tax_type__code=3, receipt=receipt)
        return receipt

    def _bad_receipt(self):
        receipt = fixtures.ReceiptFactory(
            point_of_sales=models.PointOfSales.objects.first(),
            document_type__code=80,
        )
        fixtures.VatFactory(vat_type__code=5, receipt=receipt)
        fixtures.TaxFactory(tax_type__code=3, receipt=receipt)
        return receipt

    def test_validate_empty(self):
        fixtures.ReceiptFactory()

        errs = models.Receipt.objects.none().validate()

        self.assertEqual(errs, [])
        self.assertEqual(models.ReceiptValidation.objects.count(), 0)

    def test_validation_good(self):
        """Test validating valid receipts."""
        r1 = self._good_receipt()
        r2 = self._good_receipt()
        r3 = self._good_receipt()

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            r1.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(
            r2.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(
            r3.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 3)

    def test_validation_bad(self):
        """Test validating invalid receipts."""
        self._bad_receipt()
        self._bad_receipt()
        self._bad_receipt()

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 1)
        self.assertEqual(
            errs[0],
            'Error 10015: Factura B (CbteDesde igual a CbteHasta), DocTipo: '
            '80, DocNro 203012345 no se encuentra registrado en los padrones '
            'de AFIP y no corresponde a una cuit pais.'
        )
        self.assertQuerysetEqual(models.ReceiptValidation.objects.all(), [])

    def test_validation_mixed(self):
        """
        Test validating a mixture of valid and invalid receipts.

        Receipts are validated by AFIP in-order, so all receipts previous to
        the bad one are validated, and nothing else is even parsed after the
        invalid one.
        """
        r1 = self._good_receipt()
        self._bad_receipt()
        self._good_receipt()

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 1)
        self.assertEqual(
            errs[0],
            'Error 10015: Factura B (CbteDesde igual a CbteHasta), DocTipo: '
            '80, DocNro 203012345 no se encuentra registrado en los padrones '
            'de AFIP y no corresponde a una cuit pais.'
        )
        self.assertQuerysetEqual(
            models.ReceiptValidation.objects.all(),
            [r1.pk],
            lambda rv: rv.receipt_id,
        )

    def test_validation_validated(self):
        """Test validating invalid receipts."""
        receipt = self._good_receipt()
        models.ReceiptValidation.objects.create(
            result=models.ReceiptValidation.RESULT_APPROVED,
            cae='123',
            cae_expiration=now(),
            receipt=receipt,
            processed_date=now(),
        )

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(models.ReceiptValidation.objects.count(), 1)
        self.assertEqual(errs, [])

    def test_validation_good_service(self):
        """Test validating a receipt for a service (rather than product)."""
        receipt = self._good_receipt()
        receipt.concept_id = 2
        receipt.service_start = datetime.now() - timedelta(days=10)
        receipt.service_end = datetime.now()
        receipt.expiration_date = datetime.now() + timedelta(days=10)
        receipt.save()

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            receipt.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 1)

    def test_validation_good_without_tax(self):
        """Test validating valid receipts."""
        receipt = fixtures.ReceiptFactory(
            point_of_sales=models.PointOfSales.objects.first(),
            total_amount=121,
        )
        fixtures.VatFactory(vat_type__code=5, receipt=receipt)

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            receipt.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 1)

    def test_validation_good_without_vat(self):
        """Test validating valid receipts."""
        receipt = fixtures.ReceiptFactory(
            point_of_sales=models.PointOfSales.objects.first(),
            receipt_type__code=11,
            total_amount=109,
        )
        fixtures.TaxFactory(tax_type__code=3, receipt=receipt)

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            receipt.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 1)

    def test_validation_with_observations(self):
        receipt = fixtures.ReceiptFactory(
            document_number=20291144404,
            document_type__code=80,
            point_of_sales=models.PointOfSales.objects.first(),
            receipt_type__code=1,
        )
        fixtures.VatFactory(vat_type__code=5, receipt=receipt)
        fixtures.TaxFactory(tax_type__code=3, receipt=receipt)

        errs = models.Receipt.objects.all().validate()

        self.assertEqual(len(errs), 0)
        self.assertEqual(
            receipt.validation.result,
            models.ReceiptValidation.RESULT_APPROVED,
        )
        self.assertEqual(models.ReceiptValidation.objects.count(), 1)
        self.assertEqual(models.Observation.objects.count(), 1)
        self.assertEqual(receipt.validation.observations.count(), 1)


# TODO: Test receipts with related_receipts



from unittest import mock

from django.contrib import messages
from django.contrib.admin import site
from django.contrib.auth.models import User
from django.http import HttpRequest
from django.test import Client, RequestFactory, TestCase
from django.utils.translation import ugettext as _
from factory.django import FileField

from django_afip import exceptions, models
from django_afip.admin import catch_errors, ReceiptAdmin
from testapp.testmain import fixtures


class TestCatchErrors(TestCase):

    def _get_test_instance(self, exception_type):

        class TestClass(mock.MagicMock):
            @catch_errors
            def action(self, request):
                raise exception_type

        return TestClass()

    def test_certificate_expired(self):
        obj = self._get_test_instance(exceptions.CertificateExpired)

        request = HttpRequest()
        obj.action(request)

        self.assertEqual(obj.message_user.call_count, 1)
        self.assertEqual(obj.message_user.call_args, mock.call(
            request,
            _('The AFIP Taxpayer certificate has expired.'),
            messages.ERROR,
        ))

    def test_certificate_untrusted_cert(self):
        obj = self._get_test_instance(exceptions.UntrustedCertificate)

        request = HttpRequest()
        obj.action(request)

        self.assertEqual(obj.message_user.call_count, 1)
        self.assertEqual(obj.message_user.call_args, mock.call(
            request,
            _('The AFIP Taxpayer certificate is untrusted.'),
            messages.ERROR,
        ))

    def test_certificate_auth_error(self):
        obj = self._get_test_instance(exceptions.AuthenticationError)

        request = HttpRequest()
        obj.action(request)

        self.assertEqual(obj.message_user.call_count, 1)
        self.assertEqual(obj.message_user.call_args, mock.call(
            request,
            _('An unknown authentication error has ocurred: '),
            messages.ERROR,
        ))


class TestTaxPayerAdminKeyGeneration(TestCase):

    def setUp(self):
        self.user = fixtures.SuperUserFactory()

    def test_without_key(self):
        taxpayer = fixtures.TaxPayerFactory(key=None)
        client = Client()
        client.force_login(self.user)

        response = client.post('/admin/afip/taxpayer/', data={
            '_selected_action': [taxpayer.id],
            'action': 'generate_key',
        }, follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertContains(response, 'Key generated successfully.')

        taxpayer.refresh_from_db()
        self.assertIn(
            '-----BEGIN PRIVATE KEY-----',
            taxpayer.key.file.read().decode(),
        )

    def test_with_key(self):
        taxpayer = fixtures.TaxPayerFactory(key=FileField(data=b'Blah'))
        client = Client()
        client.force_login(self.user)

        response = client.post('/admin/afip/taxpayer/', data={
            '_selected_action': [taxpayer.id],
            'action': 'generate_key',
        }, follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'No keys generated; Taxpayers already had keys.',
        )

        taxpayer.refresh_from_db()
        self.assertEqual('Blah', taxpayer.key.file.read().decode())


class TestTaxPayerAdminRequestGeneration(TestCase):

    def setUp(self):
        self.user = fixtures.SuperUserFactory()

    def test_with_csr(self):
        taxpayer = fixtures.TaxPayerFactory(key=None)
        taxpayer.generate_key()
        client = Client()
        client.force_login(self.user)

        response = client.post('/admin/afip/taxpayer/', data={
            '_selected_action': [taxpayer.id],
            'action': 'generate_csr',
        }, follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertIn(
            b'Content-Type: application/pkcs10',
            response.serialize_headers().splitlines()
        )
        self.assertContains(
            response,
            '-----BEGIN CERTIFICATE REQUEST-----',
        )

    def test_without_key(self):
        taxpayer = fixtures.TaxPayerFactory(key=None)
        taxpayer.generate_key()
        client = Client()
        client.force_login(self.user)

        response = client.post('/admin/afip/taxpayer/', data={
            '_selected_action': [taxpayer.id],
            'action': 'generate_csr',
        }, follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertIn(
            b'Content-Type: application/pkcs10',
            response.serialize_headers().splitlines()
        )
        self.assertContains(
            response,
            '-----BEGIN CERTIFICATE REQUEST-----',
        )

    def test_multiple_taxpayers(self):
        taxpayer1 = fixtures.TaxPayerFactory(key__data=b'Blah')
        taxpayer2 = fixtures.TaxPayerFactory(key__data=b'Blah')
        client = Client()
        client.force_login(self.user)

        response = client.post('/admin/afip/taxpayer/', data={
            '_selected_action': [taxpayer1.id, taxpayer2.id],
            'action': 'generate_csr',
        }, follow=True)

        self.assertEqual(response.status_code, 200)
        self.assertContains(
            response,
            'Can only generate CSR for one taxpayer at a time',
        )


class ReceiptFiltersAdminTestCase(TestCase):
    """Test ReceiptAdmin methods."""

    def setUp(self):
        fixtures.SuperUserFactory()

    def test_validation_filters(self):
        """
        Test the admin validation filters.

        This filters receipts by the validation status.
        """
        validated_receipt = fixtures.ReceiptFactory()
        failed_validation_receipt = fixtures.ReceiptFactory()
        not_validated_receipt = fixtures.ReceiptFactory()

        fixtures.ReceiptValidationFactory(receipt=validated_receipt)
        fixtures.ReceiptValidationFactory(
            result=models.ReceiptValidation.RESULT_REJECTED,
            receipt=failed_validation_receipt,
        )

        client = Client()
        client.force_login(User.objects.first())

        response = client.get('/admin/afip/receipt/?status=validated')
        self.assertContains(
            response,
            '<input class="action-select" name="_selected_action" value="{}" '
            'type="checkbox">'.format(validated_receipt.pk),
            html=True,
        )
        self.assertNotContains(
            response,
            '<input class="action-select" name="_selected_action" value="{}" '
            'type="checkbox">'.format(not_validated_receipt.pk),
            html=True,
        )
        self.assertNotContains(
            response,
            '<input class="action-select" name="_selected_action" value="{}" '
            'type="checkbox">'.format(failed_validation_receipt.pk),
            html=True,
        )

        response = client.get('/admin/afip/receipt/?status=not_validated')
        self.assertNotContains(
            response,
            '<input class="action-select" name="_selected_action" value="{}" '
            'type="checkbox">'.format(validated_receipt.pk),
            html=True,
        )
        self.assertContains(
            response,
            '<input class="action-select" name="_selected_action" value="{}" '
            'type="checkbox">'.format(not_validated_receipt.pk),
            html=True,
        )
        self.assertContains(
            response,
            '<input class="action-select" name="_selected_action" value="{}" '
            'type="checkbox">'.format(failed_validation_receipt.pk),
            html=True,
        )


class ReceiptAdminGetExcludeTestCase(TestCase):
    def test_django_111(self):
        admin = ReceiptAdmin(models.Receipt, site)
        request = RequestFactory().get('/admin/afip/receipt')
        request.user = fixtures.UserFactory()

        with mock.patch('django.VERSION', (1, 11, 7)):
            self.assertNotIn('related_receipts', admin.get_fields(request))

    def test_django_200(self):
        admin = ReceiptAdmin(models.Receipt, site)
        request = RequestFactory().get('/admin/afip/receipt')
        request.user = fixtures.UserFactory()

        with mock.patch('django.VERSION', (2, 0, 0)):
            self.assertIn('related_receipts', admin.get_fields(request))



from django.test import tag, TestCase

from django_afip import models
from testapp.testmain import fixtures


@tag('live')
class LiveAfipTestCase(TestCase):
    """
    Base class for AFIP-WS related tests.

    Since AFIP rate-limits how often authentication tokens can be fetched, we
    need to keep one between tests.
    This class is a simple hack to keep that ticket in-memory and saves it into
    the DB every time a new class is ``setUp``.
    """

    ticket = None

    def setUp(self):
        """Save a TaxPayer and Ticket into the database."""
        LiveAfipTestCase.taxpayer = fixtures.TaxPayerFactory(pk=1)

        if not LiveAfipTestCase.ticket:
            ticket = models.AuthTicket.objects.get_any_active('wsfe')
            LiveAfipTestCase.ticket = ticket

        LiveAfipTestCase.ticket.save()


class PopulatedLiveAfipTestCase(LiveAfipTestCase):
    def setUp(self):
        """Populate AFIP metadata and create a TaxPayer and PointOfSales."""
        super().setUp()
        models.populate_all()
        taxpayer = models.TaxPayer.objects.first()
        taxpayer.fetch_points_of_sales()



from datetime import date

from django.test import Client, TestCase
from django.urls import reverse

from django_afip import views
from testapp.testmain import fixtures


class ReceiptPDFTestCase(TestCase):
    def test_html_view(self):
        """Test the HTML generation view."""
        pdf = fixtures.ReceiptPDFFactory(
            receipt__concept__code=1,
            receipt__issued_date=date(2017, 5, 15),
            receipt__receipt_type__code=11,
        )
        fixtures.ReceiptValidationFactory(receipt=pdf.receipt)

        client = Client()
        response = client.get('{}?html=true'.format(
            reverse('receipt_displaypdf_view', args=(pdf.receipt.pk,))
        ))

        self.assertHTMLEqual(
            response.content.decode(),
            """
<!DOCTYPE html>
<html>
  <head>
    <meta charset="utf-8">
    <link rel="stylesheet" href="/static/receipts/receipt.css">
  </head>
  <body>
    <div class="receipt">

      <header>
        <div class="taxpayer-details group">
          <address>
            <strong>Alice Doe</strong><br>
            Happy Street 123, CABA<br>

            Responsable Monotributo<br>
          </address>

          <div class="receipt-type">
            <div class="identifier">
              C
            </div>
            <div class="code">
              Código 11
            </div>
          </div>

          <div class="receipt-details">
            <div class="receipt-type-description">
              Factura C
            </div>
            <strong>Nº</strong> None
            <br>
            <strong>Fecha de emisión:</strong> <time>May 15, 2017</time><br>
            <strong>CUIT:</strong> 20-32964233-0<br>
            <strong>Ingresos Brutos:</strong> Convenio Multilateral<br>
            <strong>Inicio de Actividades:</strong> Oct. 3, 2011
          </div>
        </div>

        <hr>

        <div class="client">
          <div><strong>Facturado a:</strong></div>
          <div class="sale-conditions">
            <strong>Condición de IVA:</strong> Consumidor Final<br>
            <strong>Condición de Pago:</strong> Contado
          </div>
          <div class="client-data">
            John Doe, DNI 203012345<br>
            La Rioja 123<br />X5000EVX Córdoba<br>
          </div>
        </div>

      </header>

      <hr>

      <table>
        <thead>
          <tr>
            <th>Descripción</th>
            <th>Cantidad</th>
            <th>Precio Unitario</th>
            <th>Monto</th>
          </tr>
        </thead>
        <tbody>

        </tbody>
        <tfoot>
          <tr>
            <td></td>
            <td></td>
            <td></td>
            <td>130.00</td>
          </tr>
        </tfoot>
      </table>

      <footer>
        <p class="cae">
          <strong>CAE</strong>
          67190616790549
          <strong>Vto CAE</strong>
          July 12, 2017
        </p>

        <div class="barcode">
          <img src="data:image/png;base64,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">
        </div>

        Consultas de validez:
        <a href="http://www.afip.gob.ar/genericos/consultacae/">
          http://www.afip.gob.ar/genericos/consultacae/
        </a>
        <br>
        Teléfono Gratuito CABA, Área de Defensa y Protección al Consumidor.
        Tel 147
      </footer>

    </div>
  </body>
</html>
            """,  # noqa: E501: It's just long stuff. :(
        )

    def test_logo_in_html(self):
        """Test the HTML generation view."""
        pdf = fixtures.ReceiptPDFFactory()
        fixtures.ReceiptValidationFactory(receipt=pdf.receipt)
        fixtures.TaxPayerExtras(taxpayer=pdf.receipt.point_of_sales.owner)

        client = Client()
        response = client.get('{}?html=true'.format(
            reverse('receipt_displaypdf_view', args=(pdf.receipt.pk,))
        ))

        self.assertContains(
            response,
            """
            <address>
            <img src="data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAAAIAAAACCAIAAAD91JpzAAAACXBIWXMAAAsTAAALEwEAmpwYAAAAB3RJTUUH4QgLERgyoZWu2QAAAB1pVFh0Q29tbWVudAAAAAAAQ3JlYXRlZCB3aXRoIEdJTVBkLmUHAAAAEklEQVQI12P4//8/AwT8//8fACnkBft7DmIIAAAAAElFTkSuQmCC" alt="Logo"><br>
            <strong>Alice Doe</strong><br>
            Happy Street 123, CABA<br>

            Responsable Monotributo<br>
            </address>
            """,  # noqa: E501: It's just long stuff. :(
            html=True,
        )

    def test_pdf_view(self):
        """
        Test the PDF generation view.
        """
        taxpayer = fixtures.TaxPayerFactory()

        fixtures.TaxPayerProfileFactory(taxpayer=taxpayer)
        pdf = fixtures.ReceiptPDFFactory(
            receipt__point_of_sales__owner=taxpayer,
        )
        fixtures.ReceiptValidationFactory(receipt=pdf.receipt)

        client = Client()
        response = client.get(
            reverse('receipt_pdf_view', args=(pdf.receipt.pk,))
        )

        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.content[:7], b'%PDF-1.')

        headers = sorted(response.serialize_headers().decode().splitlines())
        self.assertIn('Content-Type: application/pdf', headers)


class ReceiptPDFViewDownloadNameTestCase(TestCase):
    def test_download_name(self):
        fixtures.ReceiptFactory(pk=9, receipt_number=32)

        view = views.ReceiptPDFView()
        view.kwargs = {'pk': 9}

        self.assertEqual(view.get_download_name(), '0001-00000032.pdf')



import os

from barcode.writer import SVGWriter
from django.test import TestCase

from django_afip import models
from django_afip.pdf import ReceiptBarcodeGenerator
from testapp.testmain import fixtures


class ReceiptPDFGenerationTestCase(TestCase):
    def test_pdf_generation(self):
        """
        Test PDF file generation.

        For the moment, this test case mostly verifies that pdf generation
        *works*, but does not actually validate the pdf file itself.

        Running this locally *will* yield the file itself, which is useful for
        manual inspection.
        """
        pdf = fixtures.ReceiptPDFFactory(receipt__receipt_number=3)
        fixtures.ReceiptValidationFactory(receipt=pdf.receipt)
        pdf.save_pdf()
        self.assertTrue(pdf.pdf_file.name.startswith('receipts/'))
        self.assertTrue(pdf.pdf_file.name.endswith('.pdf'))

    def test_unauthorized_receipt_generation(self):
        """
        Test PDF file generation for unauthorized receipts.

        Confirm that attempting to generate a PDF for an unauthorized receipt
        raises.
        """
        taxpayer = fixtures.TaxPayerFactory()
        fixtures.TaxPayerProfileFactory(taxpayer=taxpayer)
        receipt = fixtures.ReceiptFactory(
            receipt_number=None,
            point_of_sales__owner=taxpayer,
        )
        pdf = models.ReceiptPDF.objects.create_for_receipt(
            receipt=receipt,
            client_name='John Doe',
            client_address='12 Green Road\nGreenville\nUK',
        )
        with self.assertRaisesMessage(
            Exception,
            'Cannot generate pdf for non-authorized receipt'
        ):
            pdf.save_pdf()


class BarcodeGeneratorVerificationDigitTestCase(TestCase):
    def assert_verification_digit(self, number, digit):
        number = [int(n) for n in number]
        self.assertEqual(
            ReceiptBarcodeGenerator.verification_digit(number),
            digit,
        )

    def test_verfification_digit(self):
        self.assert_verification_digit('01234567890', 5)
        self.assert_verification_digit(
            '123456789012345678901234567890123456789',
            0,
        )
        self.assert_verification_digit(
            '111111111112233334444444444444455555555',
            3,
        )
        self.assert_verification_digit(
            '202468793631100026719363618010820170512',
            9
        )


class BarcodeGeneratorTestCase(TestCase):
    def test_generate_barcode(self):
        """
        Test that barcode generation matches that from a pre-generated example.
        """
        receipt = fixtures.ReceiptFactory(
            point_of_sales__number=1,
            receipt_type__code=11,
        )
        fixtures.ReceiptValidationFactory(receipt=receipt,)

        generator = ReceiptBarcodeGenerator(receipt)
        barcode = generator.generate_barcode(SVGWriter)

        # Remove the line that stamps the pyBarcode version
        # Otherwise this test will break as soon as a new release comes out.
        lines = [
            line for line in barcode.splitlines()
            if b'Autogenerated with pyBarcode' not in line
        ]

        directory = os.path.dirname(os.path.abspath(__file__))
        path = os.path.join(directory, 'barcode.svg')

        with open(path, 'rb') as f:
            self.assertEqual(lines, f.read().splitlines())


class GenerateReceiptPDFSignalTestCase(TestCase):
    def test_not_validated_receipt(self):
        printable = fixtures.ReceiptPDFFactory()

        self.assertFalse(printable.pdf_file)

    def test_validated_receipt(self):
        validation = fixtures.ReceiptValidationFactory()
        printable = fixtures.ReceiptPDFFactory(receipt=validation.receipt)

        self.assertTrue(printable.pdf_file)
        self.assertTrue(printable.pdf_file.name.endswith('.pdf'))



import re
import warnings

from django import template

register = template.Library()


@register.filter
def receiptnumber(receipt):
    warnings.warn(
        'Use receipt.formatted_number instead.',
        DeprecationWarning,
        stacklevel=2,
    )
    return receipt.formatted_number


@register.filter
def format_cuit(cuit):
    numbers = re.sub('[^\\d]', '', str(cuit))
    if len(numbers) != 11:
        return cuit
    return '{}-{}-{}'.format(
        numbers[0:2],
        numbers[2:10],
        numbers[10:11]
    )



import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('afip', '0001_squashed_0036_receiptpdf__client_address__blank'),
    ]

    operations = [
        migrations.CreateModel(
            name='TaxPayerExtras',
            fields=[
                (
                    'id',
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    )
                ),
                (
                    'logo',
                    models.ImageField(
                        blank=True,
                        help_text=('A logo to use when generating printable '
                                   'receipts.'),
                        null=True,
                        upload_to='afip/taxpayers/logos/',
                        verbose_name='pdf file',
                    )
                ),
                (
                    'taxpayer',
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name='extras',
                        to='afip.TaxPayer',
                        verbose_name='taxpayer',
                    )
                ),
            ],
            options={
                'verbose_name': 'taxpayer extras',
                'verbose_name_plural': 'taxpayers extras',
            },
        ),
    ]



from django.core.management.base import BaseCommand
from django.utils.translation import ugettext as _

from django_afip import models


class Command(BaseCommand):
    help = _('Fetches required AFIP metadata and stores it locally')
    requires_migrations_checks = True

    def handle(self, *args, **options):
        models.populate_all()



from OpenSSL import crypto

from django_afip import exceptions

PKCS7_NOSIGS = 0x4  # defined in pkcs7.h


def create_embeded_pkcs7_signature(data, cert, key):
    """
    Creates an embeded ("nodetached") pkcs7 signature.

    This is equivalent to the output of::

        openssl smime -sign -signer cert -inkey key -outform DER -nodetach < data

    :type data: bytes
    :type cert: str
    :type key: str
    """  # noqa: E501

    assert isinstance(data, bytes)
    assert isinstance(cert, str)

    try:
        pkey = crypto.load_privatekey(crypto.FILETYPE_PEM, key)
        signcert = crypto.load_certificate(crypto.FILETYPE_PEM, cert)
    except crypto.Error as e:
        raise exceptions.CorruptCertificate from e

    bio_in = crypto._new_mem_buf(data)
    pkcs7 = crypto._lib.PKCS7_sign(
        signcert._x509, pkey._pkey, crypto._ffi.NULL, bio_in, PKCS7_NOSIGS
    )
    bio_out = crypto._new_mem_buf()
    crypto._lib.i2d_PKCS7_bio(bio_out, pkcs7)
    signed_data = crypto._bio_to_string(bio_out)

    return signed_data


def create_key(file_):
    """
    Create a key and save it into ``file_``.

    Note that ``file`` must be opened in binary mode.
    """
    pkey = crypto.PKey()
    pkey.generate_key(crypto.TYPE_RSA, 2048)

    file_.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, pkey))
    file_.flush()


def create_csr(key_file, organization_name, common_name, serial_number, file_):
    """Create a CSR for a key, and save it into ``file``."""
    key = crypto.load_privatekey(crypto.FILETYPE_PEM, key_file.read())

    req = crypto.X509Req()
    subj = req.get_subject()

    subj.O = organization_name  # noqa: E741 (we can't do anything about this)
    subj.CN = common_name
    subj.serialNumber = serial_number

    req.set_pubkey(key)
    req.sign(key, 'md5')

    file_.write(crypto.dump_certificate_request(crypto.FILETYPE_PEM, req))


def parse_certificate(file_):
    return crypto.load_certificate(crypto.FILETYPE_PEM, file_)



import os
from datetime import date, datetime

from django.conf import settings
from django.contrib.auth.models import User
from django.utils.timezone import make_aware
from factory import LazyFunction, PostGenerationMethodCall, SubFactory
from factory.django import DjangoModelFactory, FileField, ImageField

from django_afip import models


def _key_file():
    return open(os.path.join(settings.BASE_DIR, 'test.key'))


def _cert_file():
    return open(os.path.join(settings.BASE_DIR, 'test.crt'))


def _tiny_image_file():
    return open(os.path.join(settings.BASE_DIR, 'tiny.png'), 'rb')


class UserFactory(DjangoModelFactory):
    class Meta:
        model = User

    username = 'john doe'
    email = 'john@doe.co'
    password = PostGenerationMethodCall('set_password', '123')


class SuperUserFactory(UserFactory):
    is_staff = True
    is_superuser = True


class ConceptTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.ConceptType
        django_get_or_create = ('code',)

    code = '1'
    description = 'Producto'
    valid_from = date(2010, 9, 17)


class DocumentTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.DocumentType

    code = '96'
    description = 'DNI'
    valid_from = date(2008, 7, 25)


class CurrencyTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.CurrencyType

    code = 'PES'
    description = 'Pesos Argentinos'
    valid_from = date(2009, 4, 3)


class ReceiptTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.ReceiptType
        django_get_or_create = ('code',)

    code = '11'
    description = 'Factura C'
    valid_from = date(2011, 3, 30)


class TaxPayerFactory(DjangoModelFactory):
    class Meta:
        model = models.TaxPayer

    name = 'John Smith'
    cuit = 20329642330
    is_sandboxed = True
    key = FileField(from_func=_key_file)
    certificate = FileField(from_func=_cert_file)
    active_since = datetime(2011, 10, 3)


class TaxPayerProfileFactory(DjangoModelFactory):
    class Meta:
        model = models.TaxPayerProfile

    taxpayer = SubFactory(TaxPayerFactory)
    issuing_name = 'Red Company Inc.'
    issuing_address = '100 Red Av\nRedsville\nUK'
    issuing_email = 'billing@example.com'
    vat_condition = 'Exempt'
    gross_income_condition = 'Exempt'
    sales_terms = 'Credit Card'


class PointOfSalesFactory(DjangoModelFactory):
    class Meta:
        model = models.PointOfSales

    number = 1
    issuance_type = 'CAE'
    blocked = False
    owner = SubFactory(TaxPayerFactory)


class ReceiptFactory(DjangoModelFactory):
    class Meta:
        model = models.Receipt

    concept = SubFactory(ConceptTypeFactory, code=1)
    document_type = SubFactory(DocumentTypeFactory, code=96)
    document_number = 203012345
    issued_date = datetime(2017, 5, 15)
    issued_date = LazyFunction(date.today)
    total_amount = 130
    net_untaxed = 0
    net_taxed = 100
    exempt_amount = 0
    currency = SubFactory(CurrencyTypeFactory, code='PES')
    currency_quote = 1
    receipt_type = SubFactory(ReceiptTypeFactory, code=6)
    point_of_sales = SubFactory(PointOfSalesFactory)


class ReceiptValidationFactory(DjangoModelFactory):
    class Meta:
        model = models.ReceiptValidation

    result = models.ReceiptValidation.RESULT_APPROVED
    processed_date = make_aware(datetime(2017, 7, 2, 21, 6, 4))
    cae = '67190616790549'
    cae_expiration = make_aware(datetime(2017, 7, 12))
    receipt = SubFactory(ReceiptFactory, receipt_number=17)


class ReceiptPDFFactory(DjangoModelFactory):
    class Meta:
        model = models.ReceiptPDF

    client_address = 'La Rioja 123\nX5000EVX Córdoba'
    client_name = 'John Doe'
    client_vat_condition = 'Consumidor Final'
    gross_income_condition = 'Convenio Multilateral'
    issuing_address = 'Happy Street 123, CABA'
    issuing_name = 'Alice Doe'
    receipt = SubFactory(ReceiptFactory)
    sales_terms = 'Contado'
    vat_condition = 'Responsable Monotributo'


class GenericAfipTypeFactory(DjangoModelFactory):
    class Meta:
        model = models.GenericAfipType

    code = 80
    description = 'CUIT'
    valid_from = datetime(2017, 8, 10)


class VatTypeFactory(GenericAfipTypeFactory):
    class Meta:
        model = models.VatType


class TaxTypeFactory(GenericAfipTypeFactory):
    class Meta:
        model = models.TaxType


class VatFactory(DjangoModelFactory):
    class Meta:
        model = models.Vat

    amount = 21
    base_amount = 100
    receipt = SubFactory(ReceiptFactory)
    vat_type = SubFactory(VatTypeFactory)


class TaxFactory(DjangoModelFactory):
    class Meta:
        model = models.Tax

    aliquot = 9
    amount = 9
    base_amount = 100
    receipt = SubFactory(ReceiptFactory)
    tax_type = SubFactory(TaxTypeFactory)


class TaxPayerExtras(DjangoModelFactory):
    class Meta:
        model = models.TaxPayerExtras

    taxpayer = SubFactory(TaxPayerFactory)
    logo = ImageField(from_func=_tiny_image_file)



import functools
import logging
from datetime import datetime

import django
from django.apps import apps
from django.contrib import admin, messages
from django.contrib.admin.sites import AlreadyRegistered
from django.db.models import F
from django.http import HttpResponse
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.utils.translation import ugettext as _

from django_afip import exceptions, models


logger = logging.getLogger(__name__)


# TODO: Add an action to populate generic types.


def catch_errors(f):
    """
    Catches specific errors in admin actions and shows a friendly error.
    """

    @functools.wraps(f)
    def wrapper(self, request, *args, **kwargs):
        try:
            return f(self, request, *args, **kwargs)
        except exceptions.CertificateExpired as e:
            self.message_user(
                request,
                _('The AFIP Taxpayer certificate has expired.'),
                messages.ERROR,
            )
        except exceptions.UntrustedCertificate as e:
            self.message_user(
                request,
                _('The AFIP Taxpayer certificate is untrusted.'),
                messages.ERROR,
            )
        except exceptions.CorruptCertificate as e:
            self.message_user(
                request,
                _('The AFIP Taxpayer certificate is corrupt.'),
                messages.ERROR,
            )
        except exceptions.AuthenticationError as e:
            logger.exception('AFIP auth failed')
            self.message_user(
                request,
                _('An unknown authentication error has ocurred: %s') % e,
                messages.ERROR,
            )

    return wrapper


class VatInline(admin.TabularInline):
    model = models.Vat
    fields = (
        'vat_type',
        'base_amount',
        'amount',
    )
    extra = 1


class TaxInline(admin.TabularInline):
    model = models.Tax
    fields = (
        'tax_type',
        'description',
        'base_amount',
        'aliquot',
        'amount',
    )
    extra = 1


class ReceiptEntryInline(admin.TabularInline):
    model = models.ReceiptEntry
    fields = (
        'description',
        'quantity',
        'unit_price',
        'vat',
    )
    extra = 1


class ReceiptStatusFilter(admin.SimpleListFilter):
    title = _('status')
    parameter_name = 'status'

    VALIDATED = 'validated'
    NOT_VALIDATED = 'not_validated'

    def lookups(self, request, model_admin):
        return (
            (self.VALIDATED, _('Validated')),
            (self.NOT_VALIDATED, _('Not validated')),
        )

    def queryset(self, request, queryset):
        if self.value() == self.VALIDATED:
            return queryset.filter(
                validation__result=models.ReceiptValidation.RESULT_APPROVED
            )
        if self.value() == self.NOT_VALIDATED:
            return queryset.exclude(
                validation__result=models.ReceiptValidation.RESULT_APPROVED
            )


class ReceiptTypeFilter(admin.SimpleListFilter):
    title = models.ReceiptType._meta.verbose_name
    parameter_name = 'receipt_type'

    def lookups(self, request, model_admin):
        return (
            (receipt_type.code, receipt_type.description)
            for receipt_type in models.ReceiptType.objects.filter(
                receipts__isnull=False,
            ).distinct()
        )

    def queryset(self, request, queryset):
        value = self.value()
        if value:
            queryset = queryset.filter(receipt_type__code=value)
        return queryset


@admin.register(models.Receipt)
class ReceiptAdmin(admin.ModelAdmin):
    search_fields = (
        'receipt_number',
    )
    list_display = (
        'id',
        'receipt_type',
        'point_of_sales',
        'number',
        'issued_date',
        'friendly_total_amount',
        'validated',
        'pdf_link',
    )
    list_filter = (
        ReceiptStatusFilter,
        ReceiptTypeFilter,
    )
    autocomplete_fields = (
        'currency',
        'receipt_type',
        'related_receipts',
    )
    date_hierarchy = 'issued_date'

    def get_exclude(self, request, obj=None):
        if django.VERSION < (2, 0):
            # This field would load every single receipts for the widget which
            # will always result in thousands of queries until an evetual OOM.
            return ['related_receipts']
        return []

    __related_fields = [
        'validated',
        'cae',
    ]

    inlines = (
        VatInline,
        TaxInline,
        ReceiptEntryInline,
    )
    ordering = (
        '-issued_date',
    )

    readonly_fields = __related_fields

    def get_queryset(self, request):
        return super().get_queryset(request) \
            .select_related(
                'receipt_type',
                'point_of_sales',
            ).annotate(
                pdf_id=F('receiptpdf__id'),
                validation_result=F('validation__result'),
            )

    def number(self, obj):
        return obj.formatted_number
    number.short_description = _('receipt number')
    number.admin_order_field = 'receipt_number'

    def friendly_total_amount(self, obj):
        return '{:0.2f} ARS{}'.format(
            obj.total_amount * obj.currency_quote,
            '*' if obj.currency_quote != 1 else '',
        )
    friendly_total_amount.short_description = _('total amount')

    def validated(self, obj):
        return (
            obj.validation_result ==
            models.ReceiptValidation.RESULT_APPROVED
        )
    validated.short_description = _('validated')
    validated.admin_order_field = 'validation__result'
    validated.boolean = True

    def pdf_link(self, obj):
        if not obj.pdf_id:
            return mark_safe('<a href="{}?receipt={}">{}</a>'.format(
                reverse(self.admin_site.name + ':afip_receiptpdf_add'),
                obj.id,
                _('Create'),
            ))
        return mark_safe('<a href="{}">{}</a>'.format(
            reverse(
                self.admin_site.name + ':afip_receiptpdf_change',
                args=(obj.pdf_id,),
            ),
            _('Edit'),
        ))
    pdf_link.short_description = _('PDF')
    pdf_link.admin_order_field = 'receiptpdf__id'

    def cae(self, obj):
        return obj.validation.cae
    cae.short_description = _('cae')
    cae.admin_order_field = 'validation__cae'

    @catch_errors
    def validate(self, request, queryset):
        errs = queryset.validate()
        if errs:
            self.message_user(
                request,
                _('Receipt validation failed: %s') % errs,
                messages.ERROR,
            )
    validate.short_description = _('Validate')

    actions = [validate]


@admin.register(models.AuthTicket)
class AuthTicketAdmin(admin.ModelAdmin):
    list_display = (
        'unique_id',
        'owner',
        'service',
        'generated',
        'expires',
    )


@admin.register(models.TaxPayer)
class TaxPayerAdmin(admin.ModelAdmin):
    list_display = (
        'name',
        'cuit',
        'certificate_expiration',
    )

    @catch_errors
    def fetch_points_of_sales(self, request, queryset):
        poses = [
            pos
            for taxpayer in queryset.all()
            for pos in taxpayer.fetch_points_of_sales()
        ]

        created = len([pos for pos in poses if pos[1]])
        skipped = len(poses) - created

        self.message_user(
            request,
            message=(_('%d points of sales created.') % created),
            level=messages.SUCCESS,
        )
        self.message_user(
            request,
            message=(_('%d points of sales already existed.') % skipped),
            level=messages.WARNING,
        )

    fetch_points_of_sales.short_description = _('Fetch points of sales')

    def generate_key(self, request, queryset):
        key_count = sum(t.generate_key() for t in queryset.all())

        if key_count is 1:
            message = _('Key generated successfully.')
            level = messages.SUCCESS
        elif key_count is 1:
            message = _('%d keys generated successfully.') % key_count
            level = messages.SUCCESS
        else:
            message = _('No keys generated; Taxpayers already had keys.')
            level = messages.ERROR

        self.message_user(
            request,
            message=message,
            level=level,
        )
    generate_key.short_description = _('Generate key')

    def generate_csr(self, request, queryset):
        if queryset.count() > 1:
            self.message_user(
                request,
                message=_('Can only generate CSR for one taxpayer at a time'),
                level=messages.ERROR,
            )
            return

        taxpayer = queryset.first()
        if not taxpayer.key:
            taxpayer.generate_key()

        csr = taxpayer.generate_csr()
        filename = 'cuit-{}-{}.csr'.format(
            taxpayer.cuit,
            int(datetime.now().timestamp()),
        )

        response = HttpResponse(content_type='application/pkcs10')
        response['Content-Disposition'] = 'attachment; filename={}'.format(
            filename
        )

        response.write(csr.read())
        return response
    generate_csr.short_description = _('Generate CSR')

    actions = (
        fetch_points_of_sales,
        generate_key,
        generate_csr,
    )


@admin.register(models.PointOfSales)
class PointOfSalesAdmin(admin.ModelAdmin):
    list_display = (
        'owner',
        'number',
        'issuance_type',
        'blocked',
        'drop_date',
    )


@admin.register(models.CurrencyType)
class CurrencyTypeAdmin(admin.ModelAdmin):
    search_fields = (
        'code',
        'description',
    )
    list_display = (
        'code',
        'description',
        'valid_from',
        'valid_to',
    )


@admin.register(models.ReceiptType)
class ReceiptTypeAdmin(admin.ModelAdmin):
    search_fields = (
        'code',
        'description',
    )
    list_display = (
        'code',
        'description',
        'valid_from',
        'valid_to',
    )


@admin.register(models.ReceiptPDF)
class ReceiptPDFAdmin(admin.ModelAdmin):
    list_display = (
        'receipt_id',
        'client_name',
        'has_file',
    )
    raw_id_fields = (
        'receipt',
    )

    def has_file(self, obj):
        return bool(obj.pdf_file)
    has_file.admin_order_field = 'pdf_file'
    has_file.boolean = True
    has_file.short_description = _('Has file')

    def generate_pdf(self, request, queryset):
        for pdf in queryset:
            pdf.save_pdf()
    generate_pdf.short_description = _('Generate pdf')

    actions = (
        generate_pdf,
    )


@admin.register(models.ReceiptValidation)
class ReceiptValidationAdmin(admin.ModelAdmin):
    list_display = (
        'id',
        'receipt_number',
        'successful',
        'cae',
        'processed_date',
    )

    raw_id_fields = (
        'receipt',
    )

    def receipt_number(self, obj):
        return obj.receipt.formatted_number
    receipt_number.short_description = _('receipt number')
    receipt_number.admin_order_field = 'receipt_id'

    def successful(self, obj):
        return obj.result == models.ReceiptValidation.RESULT_APPROVED
    successful.short_description = _('result')
    successful.admin_order_field = 'result'
    successful.boolean = True


app = apps.get_app_config('afip')
for model in app.get_models():
    try:
        admin.site.register(model)
    except AlreadyRegistered:
        pass



import base64

from django.utils.functional import cached_property
from django_renderpdf.views import PDFView

from django_afip import models, pdf


class ReceiptPDFView(PDFView):
    @cached_property
    def receipt(self):
        return models.Receipt.objects.select_related(
            'receipt_type',
            'point_of_sales',
        ).get(
            pk=self.kwargs['pk'],
        )

    def get_download_name(self):
        return '{}.pdf'.format(self.receipt.formatted_number)

    def get_template_name(self):
        return 'receipts/code_{}.html'.format(self.receipt.receipt_type.code)

    @staticmethod
    def get_context_for_pk(pk, *args, **kwargs):
        context = {}

        receipt_pdf = models.ReceiptPDF.objects.select_related(
            'receipt',
            'receipt__receipt_type',
            'receipt__document_type',
            'receipt__validation',
            'receipt__point_of_sales',
            'receipt__point_of_sales__owner',
        ).prefetch_related(
            'receipt__entries',
        ).get(
            receipt__pk=pk,
        )

        # Prefetch required data in a single query:
        receipt_pdf.receipt = models.Receipt.objects.select_related(
            'receipt_type',
            'document_type',
            'validation',
            'point_of_sales',
            'point_of_sales__owner',
        ).prefetch_related(
            'entries',
        ).get(
            pk=receipt_pdf.receipt_id,
        )
        taxpayer = receipt_pdf.receipt.point_of_sales.owner
        extras = models.TaxPayerExtras.objects.filter(
            taxpayer=taxpayer,
        ).first()

        generator = pdf.ReceiptBarcodeGenerator(receipt_pdf.receipt)
        barcode = base64.b64encode(generator.generate_barcode())

        context['pdf'] = receipt_pdf
        context['taxpayer'] = taxpayer
        context['extras'] = extras
        context['barcode'] = barcode

        return context

    def get_context_data(self, *args, pk=None, **kwargs):
        context = super().get_context_data(*args, pk=pk, **kwargs)
        context.update(self.get_context_for_pk(pk, *args, **kwargs))
        return context


class ReceiptPDFDownloadView(ReceiptPDFView):
    prompt_download = True



#!/usr/bin/env python3

from setuptools import find_packages, setup

setup(
    name='django-afip',
    description='AFIP integration for django',
    author='Hugo Osvaldo Barrera',
    author_email='hugo@barrera.io',
    url='https://gitlab.com/WhyNotHugo/django-afip',
    license='ISC',
    packages=find_packages(),
    include_package_data=True,
    long_description=open('README.rst').read(),
    install_requires=[
        'django>=1.11',
        'django_renderpdf>=0.1.0',
        'lxml>=3.4.4',
        'pyopenssl>=16.2.0',
        'python-barcode>=0.8.0',
        'pytz>=2015.4',
        'setuptools-git>=1.1',
        'setuptools-scm>=1.7.0',
        'wheel>=0.24.0',
        'zeep>=1.1.0',
    ],
    extras_require={
        'docs': ['Sphinx', 'sphinx-autobuild'],
        'postgres': ['psycopg2'],
        'mysql': ['mysqlclient'],
    },
    use_scm_version={
        'version_scheme': 'post-release',
        'write_to': 'django_afip/version.py',
    },
    setup_requires=['setuptools_scm'],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: ISC License (ISCL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



__all__ = (
    'get_client',
)

import pytz
from zeep import Client
from zeep.cache import SqliteCache
from zeep.transports import Transport

TZ_AR = pytz.timezone(pytz.country_timezones['ar'][0])


transport = Transport(cache=SqliteCache(timeout=86400))
wsdls = {
    ('wsaa', False): 'https://wsaa.afip.gov.ar/ws/services/LoginCms?wsdl',
    ('wsfe', False): 'https://servicios1.afip.gov.ar/wsfev1/service.asmx?WSDL',
    ('wsaa', True): 'https://wsaahomo.afip.gov.ar/ws/services/LoginCms?wsdl',
    ('wsfe', True): 'https://wswhomo.afip.gov.ar/wsfev1/service.asmx?WSDL',
}

cached_clients = {}


def get_client(service_name, sandbox=False):
    """
    Returns a client for a given service.

    The `sandbox` argument should only be necessary if a the client will be
    used to make a request. If it will only be used to serialize objects, it is
    irrelevant.  Avoid the overhead of determining the sandbox mode in the
    calling context if only serialization operations will take place.

    :param string service_name: The name of the web services.
    :param bool sandbox: Whether the sandbox (or production) environment should
        be used by the returned client.
    :returns: A zeep client to communicate with an AFIP webservice.
    :rtype: zeep.Client
    """
    key = (service_name.lower(), sandbox,)

    try:
        if key not in cached_clients:
            cached_clients[key] = Client(wsdls[key], transport=transport)

        return cached_clients[key]
    except KeyError:
        raise ValueError('Unknown service name, {}'.format(service_name))



from django.apps import AppConfig


class AfipConfig(AppConfig):
    name = 'django_afip'
    label = 'afip'
    verbose_name = 'AFIP'

    def ready(self):
        # Register app signals:
        from django_afip import signals  # noqa: F401



from django.db.models import Sum

from . import clients


def _wsfe_factory():
    return clients.get_client('wsfe').type_factory('ns0')


def serialize_datetime(datetime):
    """
    "Another date formatting function?" you're thinking, eh? Well, this
    actually formats dates in the *exact* format the AFIP's WS expects it,
    which is almost like ISO8601.

    Note that .isoformat() works fine on production servers, but not on the
    sandbox ones.
    """
    return datetime.strftime('%Y-%m-%dT%H:%M:%S-00:00')


def serialize_date(date):
    return date.strftime('%Y%m%d')


def serialize_ticket(ticket):
    return _wsfe_factory().FEAuthRequest(
        Token=ticket.token,
        Sign=ticket.signature,
        Cuit=ticket.owner.cuit,
    )


def serialize_multiple_receipts(receipts):
    receipts = receipts.all().order_by('receipt_number')
    f = _wsfe_factory()

    first = receipts.first()
    receipts = [serialize_receipt(receipt) for receipt in receipts]

    wso = f.FECAERequest(
        FeCabReq=f.FECAECabRequest(
            CantReg=len(receipts),
            PtoVta=first.point_of_sales.number,
            CbteTipo=first.receipt_type.code,
        ),
        FeDetReq=f.ArrayOfFECAEDetRequest(receipts),
    )

    # for receipt in receipts:
    #     wso.FeDetReq.FECAEDetRequest.append(receipt)

    return wso


def serialize_receipt(receipt):
    from django_afip import models
    f = _wsfe_factory()
    subtotals = models.Receipt.objects.filter(pk=receipt.pk).aggregate(
        vat=Sum('vat__amount', distinct=True),
        taxes=Sum('taxes__amount', distinct=True),
    )

    serialized = f.FECAEDetRequest(
        Concepto=receipt.concept.code,
        DocTipo=receipt.document_type.code,
        DocNro=receipt.document_number,
        # TODO: Check that this is not None!,
        CbteDesde=receipt.receipt_number,
        CbteHasta=receipt.receipt_number,
        CbteFch=serialize_date(receipt.issued_date),
        ImpTotal=receipt.total_amount,
        ImpTotConc=receipt.net_untaxed,
        ImpNeto=receipt.net_taxed,
        ImpOpEx=receipt.exempt_amount,
        ImpIVA=subtotals['vat'] or 0,
        ImpTrib=subtotals['taxes'] or 0,
        MonId=receipt.currency.code,
        MonCotiz=receipt.currency_quote,
    )
    if int(receipt.concept.code) in (2, 3,):
        serialized.FchServDesde = serialize_date(receipt.service_start)
        serialized.FchServHasta = serialize_date(receipt.service_end)
        serialized.FchVtoPago = serialize_date(receipt.expiration_date)

    if receipt.taxes.count():
        serialized.Tributos = f.ArrayOfTributo([
            serialize_tax(tax) for tax in receipt.taxes.all()
        ])
    if receipt.vat.count():
        serialized.Iva = f.ArrayOfAlicIva([
            serialize_vat(vat) for vat in receipt.vat.all()
        ])

    # XXX: This was never finished!
    # serialized.CbtesAsoc = f.ArrayOfCbteAsoc([
    #     f.CbteAsoc(
    #         receipt.receipt_type.code,
    #         receipt.point_of_sales.number,
    #         receipt.receipt_number,
    #     ) for r in receipt.related_receipts.all()
    # ])

    return serialized


def serialize_tax(tax):
    return _wsfe_factory().Tributo(
        Id=tax.tax_type.code,
        Desc=tax.description,
        BaseImp=tax.base_amount,
        Alic=tax.aliquot,
        Importe=tax.amount,
    )


def serialize_vat(vat):
    return _wsfe_factory().AlicIva(
        Id=vat.vat_type.code,
        BaseImp=vat.base_amount,
        Importe=vat.amount,
    )
